from stripe.api_resources.abstract import ListableAPIResource


class BalanceTransaction(ListableAPIResource):
    OBJECT_NAME = 'balance_transaction'

    @classmethod
    def class_url(cls):
        return '/v1/balance/history'



import stripe
from stripe.test.helper import StripeResourceTest


class TransferTest(StripeResourceTest):

    def test_list_transfers(self):
        stripe.Transfer.list()
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers',
            {}
        )

    def test_cancel_transfer(self):
        transfer = stripe.Transfer(id='tr_cancel')
        transfer.cancel()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/transfers/tr_cancel/cancel',
            {},
            None
        )


class TransferReversalsTests(StripeResourceTest):
    def test_create_reversal(self):
        stripe.Transfer.create_reversal(
            'tr_123',
            amount=100
        )
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/transfers/tr_123/reversals',
            {'amount': 100},
            None
        )

    def test_retrieve_reversal(self):
        stripe.Transfer.retrieve_reversal(
            'tr_123',
            'trr_123'
        )
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers/tr_123/reversals/trr_123',
            {},
            None
        )

    def test_modify_reversal(self):
        stripe.Transfer.modify_reversal(
            'tr_123',
            'trr_123',
            metadata={'foo': 'bar'}
        )
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/transfers/tr_123/reversals/trr_123',
            {'metadata': {'foo': 'bar'}},
            None
        )

    def test_list_reversals(self):
        stripe.Transfer.list_reversals(
            'tr_123'
        )
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers/tr_123/reversals',
            {},
            None
        )



import stripe
from stripe.test.helper import StripeResourceTest


class MetadataTest(StripeResourceTest):

    def test_noop_metadata(self):
        charge = stripe.Charge(id='ch_foo')
        charge.description = 'test'
        charge.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_foo',
            {
                'description': 'test',
            },
            None
        )

    def test_unset_metadata(self):
        charge = stripe.Charge(id='ch_foo')
        charge.metadata = {}
        charge.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_foo',
            {
                'metadata': {},
            },
            None
        )

    def test_whole_update(self):
        charge = stripe.Charge(id='ch_foo')
        charge.metadata = {'whole': 'update'}
        charge.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_foo',
            {
                'metadata': {'whole': 'update'},
            },
            None
        )

    def test_individual_delete(self):
        charge = stripe.Charge(id='ch_foo')
        charge.metadata = {'whole': None}
        charge.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_foo',
            {
                'metadata': {'whole': None},
            },
            None
        )



from stripe import api_requestor, util
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class Charge(CreateableAPIResource, ListableAPIResource,
             UpdateableAPIResource):
    OBJECT_NAME = 'charge'

    def refund(self, idempotency_key=None, **params):
        url = self.instance_url() + '/refund'
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, params, headers))
        return self

    def capture(self, idempotency_key=None, **params):
        url = self.instance_url() + '/capture'
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, params, headers))
        return self

    def update_dispute(self, idempotency_key=None, **params):
        requestor = api_requestor.APIRequestor(self.api_key,
                                               api_version=self.stripe_version,
                                               account=self.stripe_account)
        url = self.instance_url() + '/dispute'
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request('post', url, params, headers)
        self.refresh_from({'dispute': response}, api_key, True)
        return self.dispute

    def close_dispute(self, idempotency_key=None):
        requestor = api_requestor.APIRequestor(self.api_key,
                                               api_version=self.stripe_version,
                                               account=self.stripe_account)
        url = self.instance_url() + '/dispute/close'
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request('post', url, {}, headers)
        self.refresh_from({'dispute': response}, api_key, True)
        return self.dispute

    def mark_as_fraudulent(self, idempotency_key=None):
        params = {
            'fraud_details': {'user_report': 'fraudulent'}
        }
        url = self.instance_url()
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, params, headers))
        return self

    def mark_as_safe(self, idempotency_key=None):
        params = {
            'fraud_details': {'user_report': 'safe'}
        }
        url = self.instance_url()
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, params, headers))
        return self



from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class Product(CreateableAPIResource, UpdateableAPIResource,
              ListableAPIResource, DeletableAPIResource):
    OBJECT_NAME = 'product'



import urllib
import warnings

from stripe import api_requestor, util
from stripe.api_resources.charge import Charge
from stripe.api_resources.invoice import Invoice
from stripe.api_resources.invoice_item import InvoiceItem
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import nested_resource_class_methods


@nested_resource_class_methods(
    'source',
    operations=['create', 'retrieve', 'update', 'delete', 'list']
)
class Customer(CreateableAPIResource, UpdateableAPIResource,
               ListableAPIResource, DeletableAPIResource):
    OBJECT_NAME = 'customer'

    def add_invoice_item(self, idempotency_key=None, **params):
        params['customer'] = self.id
        ii = InvoiceItem.create(self.api_key,
                                idempotency_key=idempotency_key, **params)
        return ii

    def invoices(self, **params):
        params['customer'] = self.id
        invoices = Invoice.list(self.api_key, **params)
        return invoices

    def invoice_items(self, **params):
        params['customer'] = self.id
        iis = InvoiceItem.list(self.api_key, **params)
        return iis

    def charges(self, **params):
        params['customer'] = self.id
        charges = Charge.list(self.api_key, **params)
        return charges

    def update_subscription(self, idempotency_key=None, **params):
        warnings.warn(
            'The `update_subscription` method is deprecated. Instead, use the '
            '`subscriptions` resource on the customer object to update a '
            'subscription',
            DeprecationWarning)
        requestor = api_requestor.APIRequestor(self.api_key,
                                               api_version=self.stripe_version,
                                               account=self.stripe_account)
        url = self.instance_url() + '/subscription'
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request('post', url, params, headers)
        self.refresh_from({'subscription': response}, api_key, True)
        return self.subscription

    def cancel_subscription(self, idempotency_key=None, **params):
        warnings.warn(
            'The `cancel_subscription` method is deprecated. Instead, use the '
            '`subscriptions` resource on the customer object to cancel a '
            'subscription',
            DeprecationWarning)
        requestor = api_requestor.APIRequestor(self.api_key,
                                               api_version=self.stripe_version,
                                               account=self.stripe_account)
        url = self.instance_url() + '/subscription'
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request('delete', url, params, headers)
        self.refresh_from({'subscription': response}, api_key, True)
        return self.subscription

    # TODO: Remove arg in next major release.
    def delete_discount(self, **params):
        requestor = api_requestor.APIRequestor(self.api_key,
                                               api_version=self.stripe_version,
                                               account=self.stripe_account)
        url = self.instance_url() + '/discount'
        _, api_key = requestor.request('delete', url)
        self.refresh_from({'discount': None}, api_key, True)

    @classmethod
    def modify_source(cls, sid, source_id, **params):
        url = "%s/%s/sources/%s" % (
            cls.class_url(), urllib.quote_plus(util.utf8(sid)),
            urllib.quote_plus(util.utf8(source_id)))
        return cls._modify(url, **params)



import warnings

from stripe import api_requestor, util
from stripe.api_resources.abstract import DeletableAPIResource


class EphemeralKey(DeletableAPIResource):
    OBJECT_NAME = 'ephemeral_key'

    @classmethod
    def class_name(cls):
        return 'ephemeral_key'

    @classmethod
    def create(cls, api_key=None, idempotency_key=None,
               stripe_version=None, stripe_account=None,
               api_version=None, **params):
        if stripe_version is None:
            if api_version is not None:
                stripe_version = api_version
                warnings.warn(
                    "The `api_version` parameter when creating an ephemeral "
                    "key is deprecated. Please use `stripe_version` instead.",
                    DeprecationWarning)
            else:
                raise ValueError(
                    "stripe_version must be specified to create an ephemeral "
                    "key")

        requestor = api_requestor.APIRequestor(
            api_key,
            api_version=stripe_version,
            account=stripe_account
        )

        url = cls.class_url()
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request('post', url, params, headers)
        return util.convert_to_stripe_object(response, api_key, stripe_version,
                                             stripe_account)



from stripe.api_resources.abstract import ListableAPIResource


class OrderReturn(ListableAPIResource):
    OBJECT_NAME = 'order_return'

    @classmethod
    def class_url(cls):
        return '/v1/order_returns'



import stripe
from stripe.test.helper import StripeResourceTest


class RefundTest(StripeResourceTest):

    def test_create_refund(self):
        stripe.Refund.create(charge='ch_foo')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/refunds',
            {'charge': 'ch_foo'},
            None
        )

    def test_fetch_refund(self):
        stripe.Refund.retrieve('re_foo')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/refunds/re_foo',
            {},
            None
        )

    def test_list_refunds(self):
        stripe.Refund.list(limit=3, charge='ch_foo')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/refunds',
            {'limit': 3, 'charge': 'ch_foo'}
        )

    def test_update_refund(self):
        refund = stripe.Refund.construct_from({
            'id': "ref_update",
            'charge': "ch_update",
            'metadata': {},
        }, 'api_key')
        refund.metadata["key"] = "value"
        refund.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/refunds/ref_update',
            {
                'metadata': {
                    'key': 'value',
                }
            },
            None
        )


class ChargeRefundTest(StripeResourceTest):

    def test_create_refund(self):
        charge = stripe.Charge.construct_from({
            'id': 'ch_foo',
            'refunds': {
                'object': 'list',
                'url': '/v1/charges/ch_foo/refunds',
            }
        }, 'api_key')

        charge.refunds.create()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_foo/refunds',
            {},
            None
        )

    def test_non_recursive_save(self):
        charge = stripe.Charge.construct_from({
            'id': 'ch_nested_update',
            'customer': {
                'object': 'customer',
                'description': 'foo',
            },
            'refunds': {
                'object': 'list',
                'url': '/v1/charges/ch_foo/refunds',
                'data': [{
                    'id': 'ref_123',
                }],
            },
        }, 'api_key')

        charge.customer.description = 'bar'
        charge.refunds.has_more = True
        charge.refunds.data[0].description = 'bar'
        charge.save()

        # Note: ideally, we'd want the library to NOT issue requests in this
        # case (i.e. the assert should actually be `assert_not_called()`).
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_nested_update',
            {'refunds': {'has_more': True}},
            None
        )

    def test_fetch_refund(self):
        charge = stripe.Charge.construct_from({
            'id': 'ch_get_refund',
            'refunds': {
                'object': 'list',
                'url': '/v1/charges/ch_get_refund/refunds',
            }
        }, 'api_key')

        charge.refunds.retrieve("ref_get")

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/charges/ch_get_refund/refunds/ref_get',
            {},
            None
        )

    def test_list_refunds(self):
        charge = stripe.Charge.construct_from({
            'id': 'ch_get_refund',
            'refunds': {
                'object': 'list',
                'url': '/v1/charges/ch_get_refund/refunds',
            }
        }, 'api_key')

        charge.refunds.list()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/charges/ch_get_refund/refunds',
            {},
            None
        )

    def test_update_refund(self):
        refund = stripe.Refund.construct_from({
            'id': "ref_update",
            'charge': "ch_update",
            'metadata': {},
        }, 'api_key')
        refund.metadata["key"] = "value"
        refund.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/refunds/ref_update',
            {
                'metadata': {
                    'key': 'value',
                }
            },
            None
        )



from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class Payout(CreateableAPIResource, UpdateableAPIResource,
             ListableAPIResource):
    OBJECT_NAME = 'payout'

    def cancel(self):
        self.refresh_from(self.request('post',
                          self.instance_url() + '/cancel'))



from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import nested_resource_class_methods


@nested_resource_class_methods('reversal', operations=['create', 'retrieve',
                                                       'update', 'list'])
class Transfer(CreateableAPIResource, UpdateableAPIResource,
               ListableAPIResource):
    OBJECT_NAME = 'transfer'

    def cancel(self):
        self.refresh_from(self.request('post',
                          self.instance_url() + '/cancel'))



import urllib

from stripe import util
from stripe.api_resources.customer import Customer
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class BitcoinReceiver(CreateableAPIResource, UpdateableAPIResource,
                      DeletableAPIResource, ListableAPIResource):
    OBJECT_NAME = 'bitcoin_receiver'

    def instance_url(self):
        token = util.utf8(self.id)
        extn = urllib.quote_plus(token)

        if hasattr(self, 'customer'):
            customer = util.utf8(self.customer)
            base = Customer.class_url()
            cust_extn = urllib.quote_plus(customer)
            return "%s/%s/sources/%s" % (base, cust_extn, extn)
        else:
            base = BitcoinReceiver.class_url()
            return "%s/%s" % (base, extn)

    @classmethod
    def class_url(cls):
        return '/v1/bitcoin/receivers'



from stripe import util
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class Dispute(CreateableAPIResource, ListableAPIResource,
              UpdateableAPIResource):
    OBJECT_NAME = 'dispute'

    def close(self, idempotency_key=None):
        url = self.instance_url() + '/close'
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, {}, headers))
        return self



import urllib

from stripe import util
from stripe.api_resources.transfer import Transfer
from stripe.api_resources.abstract import UpdateableAPIResource


class Reversal(UpdateableAPIResource):
    OBJECT_NAME = 'transfer_reversal'

    def instance_url(self):
        token = util.utf8(self.id)
        transfer = util.utf8(self.transfer)
        base = Transfer.class_url()
        cust_extn = urllib.quote_plus(transfer)
        extn = urllib.quote_plus(token)
        return "%s/%s/reversals/%s" % (base, cust_extn, extn)

    @classmethod
    def modify(cls, sid, **params):
        raise NotImplementedError(
            "Can't modify a reversal without a transfer"
            "ID. Call save on transfer.reversals.retrieve('reversal_id')")

    @classmethod
    def retrieve(cls, id, api_key=None, **params):
        raise NotImplementedError(
            "Can't retrieve a reversal without a transfer"
            "ID. Use transfer.reversals.retrieve('reversal_id')")



from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class ApplePayDomain(CreateableAPIResource, ListableAPIResource,
                     DeletableAPIResource):
    OBJECT_NAME = 'apple_pay_domain'

    @classmethod
    def class_url(cls):
        return '/v1/apple_pay/domains'



from stripe.api_resources.abstract import CreateableAPIResource


class ThreeDSecure(CreateableAPIResource):
    OBJECT_NAME = 'three_d_secure'

    @classmethod
    def class_url(cls):
        return '/v1/3d_secure'



import hmac
import time
from hashlib import sha256

import stripe
from stripe import error, util


class Webhook(object):
    DEFAULT_TOLERANCE = 300

    @staticmethod
    def construct_event(payload, sig_header, secret,
                        tolerance=DEFAULT_TOLERANCE, api_key=None):
        if hasattr(payload, 'decode'):
            payload = payload.decode('utf-8')
        if api_key is None:
            api_key = stripe.api_key
        data = util.json.loads(payload)
        event = stripe.Event.construct_from(data, api_key)

        WebhookSignature.verify_header(payload, sig_header, secret, tolerance)

        return event


class WebhookSignature(object):
    EXPECTED_SCHEME = 'v1'

    @staticmethod
    def _compute_signature(payload, secret):
        mac = hmac.new(secret.encode('utf-8'), msg=payload.encode('utf-8'),
                       digestmod=sha256)
        return mac.hexdigest()

    @staticmethod
    def _get_timestamp_and_signatures(header, scheme):
        list_items = map(lambda i: i.split('=', 2), header.split(','))
        timestamp = int([i[1] for i in list_items if i[0] == 't'][0])
        signatures = [i[1] for i in list_items if i[0] == scheme]
        return timestamp, signatures

    @classmethod
    def verify_header(cls, payload, header, secret, tolerance=None):
        try:
            timestamp, signatures = cls._get_timestamp_and_signatures(
                header, cls.EXPECTED_SCHEME)
        except:
            raise error.SignatureVerificationError(
                "Unable to extract timestamp and signatures from header",
                header, payload)

        if len(signatures) == 0:
            raise error.SignatureVerificationError(
                "No signatures found with expected scheme "
                "%s" % cls.EXPECTED_SCHEME,
                header, payload)

        signed_payload = "%d.%s" % (timestamp, payload)
        expected_sig = cls._compute_signature(signed_payload, secret)
        if not any(util.secure_compare(expected_sig, s) for s in signatures):
            raise error.SignatureVerificationError(
                "No signatures found matching the expected signature for "
                "payload",
                header, payload)

        if tolerance and timestamp < time.time() - tolerance:
            raise error.SignatureVerificationError(
                "Timestamp outside the tolerance zone (%d)" % timestamp,
                header, payload)

        return True



from stripe import util


class VerifyMixin(object):

    def verify(self, idempotency_key=None, **params):
        url = self.instance_url() + '/verify'
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request('post', url, params, headers))
        return self



from stripe.api_resources.abstract.api_resource import APIResource


class DeletableAPIResource(APIResource):

    def delete(self, **params):
        self.refresh_from(self.request('delete', self.instance_url(), params))
        return self



import stripe
from stripe import api_requestor, util
from stripe.api_resources.abstract import ListableAPIResource


class FileUpload(ListableAPIResource):
    OBJECT_NAME = 'file_upload'

    @classmethod
    def api_base(cls):
        return stripe.upload_api_base

    @classmethod
    def class_name(cls):
        return 'file'

    @classmethod
    def create(cls, api_key=None, api_version=None, stripe_account=None,
               **params):
        requestor = api_requestor.APIRequestor(
            api_key, api_base=cls.api_base(), api_version=api_version,
            account=stripe_account)
        url = cls.class_url()
        supplied_headers = {
            "Content-Type": "multipart/form-data"
        }
        response, api_key = requestor.request(
            'post', url, params=params, headers=supplied_headers)
        return util.convert_to_stripe_object(response, api_key, api_version,
                                             stripe_account)



import warnings

import stripe
from stripe.test.helper import StripeResourceTest


class EphemeralKeyTest(StripeResourceTest):

    def test_create(self):
        stripe.EphemeralKey.create(customer='cus_123',
                                   stripe_version='2017-05-25')

        self.requestor_class_mock.assert_called_with(
            None,
            api_version='2017-05-25',
            account=None
        )

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/ephemeral_keys',
            {'customer': 'cus_123'},
            None
        )

    def test_create_legacy_parameter_api_version(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')

            stripe.EphemeralKey.create(customer='cus_123',
                                       api_version='2017-05-25')

            self.assertEqual(1, len(w))
            self.assertEqual(w[0].category, DeprecationWarning)

        self.requestor_class_mock.assert_called_with(
            None,
            api_version='2017-05-25',
            account=None
        )

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/ephemeral_keys',
            {'customer': 'cus_123'},
            None
        )

    def test_create_without_global_version(self):
        self._do_version_tests(None)

    def test_create_with_global_version(self):
        self._do_version_tests('2017-05-25')

    def test_delete(self):
        stripe.EphemeralKey(id='ephkey_123').delete()
        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/ephemeral_keys/ephkey_123',
            {},
            None
        )

    def _do_version_tests(self, version):
        old_api_version = stripe.api_version
        stripe.api_version = version

        try:
            with self.assertRaisesRegex(ValueError,
                                        'stripe_version must be specified'):
                stripe.EphemeralKey.create(customer='cus_123')

            stripe.EphemeralKey.create(customer='cus_123',
                                       stripe_version='2017-06-05')

            self.requestor_class_mock.assert_called_with(
                None,
                api_version='2017-06-05',
                account=None
            )
        finally:
            stripe.api_version = old_api_version



import stripe
from stripe.test.helper import (
    StripeResourceTest, NOW, DUMMY_CHARGE
)


class ChargeTest(StripeResourceTest):

    def test_charge_list(self):
        stripe.Charge.list(created={'lt': NOW})

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/charges',
            {
                'created': {'lt': NOW},
            }
        )

    def test_charge_create(self):
        stripe.Charge.create(idempotency_key='foo', **DUMMY_CHARGE)

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges',
            DUMMY_CHARGE,
            {'Idempotency-Key': 'foo'},
        )

    def test_charge_retrieve(self):
        stripe.Charge.retrieve('ch_test_id')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/charges/ch_test_id',
            {},
            None
        )

    def test_charge_modify(self):
        stripe.Charge.modify('ch_test_id', refund=True)

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_test_id',
            {
                'refund': True,
            },
            None
        )

    def test_charge_update_dispute(self):
        charge = stripe.Charge(id='ch_update_id')
        charge.update_dispute(idempotency_key='foo')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_update_id/dispute',
            {},
            {'Idempotency-Key': 'foo'},
        )

    def test_charge_close_dispute(self):
        charge = stripe.Charge(id='ch_update_id')
        charge.close_dispute(idempotency_key='foo')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_update_id/dispute/close',
            {},
            {'Idempotency-Key': 'foo'},
        )

    def test_mark_as_fraudulent(self):
        charge = stripe.Charge(id='ch_update_id')
        charge.mark_as_fraudulent(idempotency_key='foo')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_update_id',
            {
                'fraud_details': {'user_report': 'fraudulent'}
            },
            {'Idempotency-Key': 'foo'},
        )

    def test_mark_as_safe(self):
        charge = stripe.Charge(id='ch_update_id')
        charge.mark_as_safe(idempotency_key='foo')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges/ch_update_id',
            {
                'fraud_details': {'user_report': 'safe'}
            },
            {'Idempotency-Key': 'foo'},
        )

    def test_create_with_source_param(self):
        stripe.Charge.create(amount=100, currency='usd',
                             source='btcrcv_test_receiver')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/charges',
            {
                'amount': 100,
                'currency': 'usd',
                'source': 'btcrcv_test_receiver'
            },
            None,
        )



import pickle
import sys
from copy import copy, deepcopy

import stripe
from stripe import util
from stripe.test.helper import StripeUnitTestCase, SAMPLE_INVOICE


class StripeObjectTests(StripeUnitTestCase):

    def test_initializes_with_parameters(self):
        obj = stripe.stripe_object.StripeObject(
            'foo', 'bar', myparam=5, yourparam='boo')

        self.assertEqual('foo', obj.id)
        self.assertEqual('bar', obj.api_key)

    def test_access(self):
        obj = stripe.stripe_object.StripeObject('myid', 'mykey', myparam=5)

        # Empty
        self.assertRaises(AttributeError, getattr, obj, 'myattr')
        self.assertRaises(KeyError, obj.__getitem__, 'myattr')
        self.assertEqual('def', obj.get('myattr', 'def'))
        self.assertEqual(None, obj.get('myattr'))

        # Setters
        obj.myattr = 'myval'
        obj['myitem'] = 'itval'
        self.assertEqual('sdef', obj.setdefault('mydef', 'sdef'))

        # Getters
        self.assertEqual('myval', obj.setdefault('myattr', 'sdef'))
        self.assertEqual('myval', obj.myattr)
        self.assertEqual('myval', obj['myattr'])
        self.assertEqual('myval', obj.get('myattr'))

        self.assertEqual(['id', 'myattr', 'mydef', 'myitem'],
                         sorted(obj.keys()))
        self.assertEqual(['itval', 'myid', 'myval', 'sdef'],
                         sorted(obj.values()))

        # Illegal operations
        self.assertRaises(ValueError, setattr, obj, 'foo', '')

    def test_refresh_from(self):
        obj = stripe.stripe_object.StripeObject.construct_from({
            'foo': 'bar',
            'trans': 'me',
        }, 'mykey')

        self.assertEqual('mykey', obj.api_key)
        self.assertEqual('bar', obj.foo)
        self.assertEqual('me', obj['trans'])
        self.assertEqual(None, obj.stripe_version)
        self.assertEqual(None, obj.stripe_account)

        obj.refresh_from({
            'foo': 'baz',
            'johnny': 5,
        }, 'key2', stripe_version='2017-08-15', stripe_account='acct_foo')

        self.assertEqual(5, obj.johnny)
        self.assertEqual('baz', obj.foo)
        self.assertRaises(AttributeError, getattr, obj, 'trans')
        self.assertEqual('key2', obj.api_key)
        self.assertEqual('2017-08-15', obj.stripe_version)
        self.assertEqual('acct_foo', obj.stripe_account)

        obj.refresh_from({
            'trans': 4,
            'metadata': {'amount': 42}
        }, 'key2', True)

        self.assertEqual('baz', obj.foo)
        self.assertEqual(4, obj.trans)

    def test_passing_nested_refresh(self):
        obj = stripe.stripe_object.StripeObject.construct_from({
            'foos': {
                'type': 'list',
                'data': [
                    {'id': 'nested'}
                ],
            }
        }, 'key', stripe_account='acct_foo')

        nested = obj.foos.data[0]

        self.assertEqual('key', obj.api_key)
        self.assertEqual('nested', nested.id)
        self.assertEqual('key', nested.api_key)
        self.assertEqual('acct_foo', nested.stripe_account)

    def test_refresh_from_nested_object(self):
        obj = stripe.stripe_object.StripeObject.construct_from(
            SAMPLE_INVOICE, 'key')

        self.assertEqual(1, len(obj.lines.subscriptions))
        self.assertTrue(
            isinstance(obj.lines.subscriptions[0],
                       stripe.stripe_object.StripeObject))
        self.assertEqual('month', obj.lines.subscriptions[0].plan.interval)

    def test_to_json(self):
        obj = stripe.stripe_object.StripeObject.construct_from(
            SAMPLE_INVOICE, 'key')

        self.check_invoice_data(util.json.loads(str(obj)))

    def check_invoice_data(self, data):
        # Check rough structure
        self.assertEqual(20, len(data.keys()))
        self.assertEqual(3, len(data['lines'].keys()))
        self.assertEqual(0, len(data['lines']['invoiceitems']))
        self.assertEqual(1, len(data['lines']['subscriptions']))

        # Check various data types
        self.assertEqual(1338238728, data['date'])
        self.assertEqual(None, data['next_payment_attempt'])
        self.assertEqual(False, data['livemode'])
        self.assertEqual('month',
                         data['lines']['subscriptions'][0]['plan']['interval'])

    def test_repr(self):
        obj = stripe.stripe_object.StripeObject(
            'foo', 'bar', myparam=5)

        obj['object'] = u'\u4e00boo\u1f00'

        res = repr(obj)

        if sys.version_info[0] < 3:
            res = unicode(repr(obj), 'utf-8')

        self.assertTrue(u'<StripeObject \u4e00boo\u1f00' in res)
        self.assertTrue(u'id=foo' in res)

    def test_pickling(self):
        obj = stripe.stripe_object.StripeObject(
            'foo', 'bar', myparam=5)

        obj['object'] = 'boo'
        obj.refresh_from(
            {
                'fala': 'lalala',
                # ensures that empty strings are correctly unpickled in Py3
                'emptystring': '',
            },
            api_key='bar', partial=True
        )

        self.assertEqual('lalala', obj.fala)

        pickled = pickle.dumps(obj)
        newobj = pickle.loads(pickled)

        self.assertEqual('foo', newobj.id)
        self.assertEqual('bar', newobj.api_key)
        self.assertEqual('boo', newobj['object'])
        self.assertEqual('lalala', newobj.fala)
        self.assertEqual('', newobj.emptystring)

    def test_deletion(self):
        obj = stripe.stripe_object.StripeObject('id', 'key')

        obj.coupon = "foo"
        self.assertEqual('foo', obj.coupon)

        del obj.coupon
        self.assertRaises(AttributeError, getattr, obj, 'coupon')

        obj.refresh_from({'coupon': 'foo'}, api_key='bar', partial=True)
        self.assertEqual('foo', obj.coupon)

    def test_copy(self):
        nested = stripe.stripe_object.StripeObject.construct_from({
            'value': 'bar',
        }, 'mykey')
        obj = stripe.stripe_object.StripeObject.construct_from({
            'empty': '',
            'value': 'foo',
            'nested': nested,
        }, 'mykey', stripe_account='myaccount')

        copied = copy(obj)

        self.assertEqual('', copied.empty)
        self.assertEqual('foo', copied.value)
        self.assertEqual('bar', copied.nested.value)

        self.assertEqual('mykey', copied.api_key)
        self.assertEqual('myaccount', copied.stripe_account)

        # Verify that we're not deep copying nested values.
        self.assertEqual(id(nested), id(copied.nested))

    def test_deepcopy(self):
        nested = stripe.stripe_object.StripeObject.construct_from({
            'value': 'bar',
        }, 'mykey')
        obj = stripe.stripe_object.StripeObject.construct_from({
            'empty': '',
            'value': 'foo',
            'nested': nested,
        }, 'mykey', stripe_account='myaccount')

        copied = deepcopy(obj)

        self.assertEqual('', copied.empty)
        self.assertEqual('foo', copied.value)
        self.assertEqual('bar', copied.nested.value)

        self.assertEqual('mykey', copied.api_key)
        self.assertEqual('myaccount', copied.stripe_account)

        # Verify that we're actually deep copying nested values.
        self.assertNotEqual(id(nested), id(copied.nested))



import urllib

from stripe import oauth, util
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import nested_resource_class_methods


@nested_resource_class_methods(
    'external_account',
    operations=['create', 'retrieve', 'update', 'delete', 'list']
)
@nested_resource_class_methods('login_link', operations=['create'])
class Account(CreateableAPIResource, ListableAPIResource,
              UpdateableAPIResource, DeletableAPIResource):
    OBJECT_NAME = 'account'

    @classmethod
    def retrieve(cls, id=None, api_key=None, **params):
        instance = cls(id, api_key, **params)
        instance.refresh()
        return instance

    @classmethod
    def modify(cls, id=None, **params):
        return cls._modify(cls._build_instance_url(id), **params)

    @classmethod
    def _build_instance_url(cls, sid):
        if not sid:
            return "/v1/account"
        sid = util.utf8(sid)
        base = cls.class_url()
        extn = urllib.quote_plus(sid)
        return "%s/%s" % (base, extn)

    def instance_url(self):
        return self._build_instance_url(self.get('id'))

    def reject(self, reason=None, idempotency_key=None):
        url = self.instance_url() + '/reject'
        headers = util.populate_headers(idempotency_key)
        if reason:
            params = {"reason": reason}
        else:
            params = {}
        self.refresh_from(
            self.request('post', url, params, headers)
        )
        return self

    def deauthorize(self, **params):
        params['stripe_user_id'] = self.id
        return oauth.OAuth.deauthorize(**params)

    @classmethod
    def modify_external_account(cls, sid, external_account_id, **params):
        url = "%s/%s/external_accounts/%s" % (
            cls.class_url(), urllib.quote_plus(util.utf8(sid)),
            urllib.quote_plus(util.utf8(external_account_id)))
        return cls._modify(url, **params)



import urllib

from stripe import error, util
from stripe.api_resources.account import Account
from stripe.api_resources.customer import Customer
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import VerifyMixin


class BankAccount(UpdateableAPIResource, DeletableAPIResource, VerifyMixin):
    OBJECT_NAME = 'bank_account'

    def instance_url(self):
        token = util.utf8(self.id)
        extn = urllib.quote_plus(token)
        if hasattr(self, 'customer'):
            customer = util.utf8(self.customer)

            base = Customer.class_url()
            owner_extn = urllib.quote_plus(customer)
            class_base = "sources"

        elif hasattr(self, 'account'):
            account = util.utf8(self.account)

            base = Account.class_url()
            owner_extn = urllib.quote_plus(account)
            class_base = "external_accounts"

        else:
            raise error.InvalidRequestError(
                "Could not determine whether bank_account_id %s is "
                "attached to a customer or an account." % token, 'id')

        return "%s/%s/%s/%s" % (base, owner_extn, class_base, extn)

    @classmethod
    def modify(cls, sid, **params):
        raise NotImplementedError(
            "Can't modify a bank account without a customer or account ID. "
            "Call save on customer.sources.retrieve('bank_account_id') or "
            "account.external_accounts.retrieve('bank_account_id') instead.")

    @classmethod
    def retrieve(cls, id, api_key=None, stripe_version=None,
                 stripe_account=None, **params):
        raise NotImplementedError(
            "Can't retrieve a bank account without a customer or account ID. "
            "Use customer.sources.retrieve('bank_account_id') or "
            "account.external_accounts.retrieve('bank_account_id') instead.")



from __future__ import print_function

import hmac
import io
import logging
import sys
import os
import re

import stripe


STRIPE_LOG = os.environ.get('STRIPE_LOG')

logger = logging.getLogger('stripe')

__all__ = [
    'io',
    'parse_qsl',
    'json',
    'utf8',
    'log_info',
    'log_debug',
    'dashboard_link',
    'logfmt',
]

try:
    from urlparse import parse_qsl
except ImportError:
    # Python < 2.6
    from cgi import parse_qsl

try:
    import json
except ImportError:
    json = None

if not (json and hasattr(json, 'loads')):
    try:
        import simplejson as json
    except ImportError:
        if not json:
            raise ImportError(
                "Stripe requires a JSON library, such as simplejson. "
                "HINT: Try installing the "
                "python simplejson library via 'pip install simplejson' or "
                "'easy_install simplejson', or contact support@stripe.com "
                "with questions.")
        else:
            raise ImportError(
                "Stripe requires a JSON library with the same interface as "
                "the Python 2.6 'json' library.  You appear to have a 'json' "
                "library with a different interface.  Please install "
                "the simplejson library.  HINT: Try installing the "
                "python simplejson library via 'pip install simplejson' "
                "or 'easy_install simplejson', or contact support@stripe.com"
                "with questions.")


def utf8(value):
    # Note the ordering of these conditionals: `unicode` isn't a symbol in
    # Python 3 so make sure to check version before trying to use it. Python
    # 2to3 will also boil out `unicode`.
    if sys.version_info < (3, 0) and isinstance(value, unicode):
        return value.encode('utf-8')
    else:
        return value


def is_appengine_dev():
    return ('APPENGINE_RUNTIME' in os.environ and
            'Dev' in os.environ.get('SERVER_SOFTWARE', ''))


def _console_log_level():
    if stripe.log in ['debug', 'info']:
        return stripe.log
    elif STRIPE_LOG in ['debug', 'info']:
        return STRIPE_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == 'debug':
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ['debug', 'info']:
        print(msg, file=sys.stderr)
    logger.info(msg)


def _test_or_live_environment():
    if stripe.api_key is None:
        return
    match = re.match(r'sk_(live|test)_', stripe.api_key)
    if match is None:
        return
    return match.groups()[0]


def dashboard_link(request_id):
    return 'https://dashboard.stripe.com/{env}/logs/{reqid}'.format(
        env=_test_or_live_environment() or 'test',
        reqid=request_id,
    )


def logfmt(props):
    def fmt(key, val):
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, basestring):
            val = unicode(val)
        if re.search(r'\s', val):
            val = repr(val)
        # key should already be a string
        if re.search(r'\s', key):
            key = repr(key)
        return u'{key}={val}'.format(key=key, val=val)
    return u' '.join([fmt(key, val) for key, val in sorted(props.items())])


# Borrowed from Django's source code
if hasattr(hmac, 'compare_digest'):
    # Prefer the stdlib implementation, when available.
    def secure_compare(val1, val2):
        return hmac.compare_digest(utf8(val1), utf8(val2))
else:
    def secure_compare(val1, val2):
        """
        Returns True if the two strings are equal, False otherwise.
        The time taken is independent of the number of characters that match.
        For the sake of simplicity, this function executes in constant time
        only when the two strings have the same length. It short-circuits when
        they have different lengths.
        """
        val1, val2 = utf8(val1), utf8(val2)
        if len(val1) != len(val2):
            return False
        result = 0
        if (sys.version_info[0] == 3 and isinstance(val1, bytes) and
                isinstance(val2, bytes)):
            for x, y in zip(val1, val2):
                result |= x ^ y
        else:
            for x, y in zip(val1, val2):
                result |= ord(x) ^ ord(y)
        return result == 0


OBJECT_CLASSES = {}


def load_object_classes():
    # This is here to avoid a circular dependency
    from stripe import api_resources

    global OBJECT_CLASSES

    OBJECT_CLASSES = {
        # data structures
        api_resources.ListObject.OBJECT_NAME: api_resources.ListObject,

        # business objects
        api_resources.Account.OBJECT_NAME: api_resources.Account,
        api_resources.AlipayAccount.OBJECT_NAME: api_resources.AlipayAccount,
        api_resources.ApplePayDomain.OBJECT_NAME: api_resources.ApplePayDomain,
        api_resources.ApplicationFee.OBJECT_NAME: api_resources.ApplicationFee,
        api_resources.ApplicationFeeRefund.OBJECT_NAME:
            api_resources.ApplicationFeeRefund,
        api_resources.Balance.OBJECT_NAME: api_resources.Balance,
        api_resources.BankAccount.OBJECT_NAME: api_resources.BankAccount,
        api_resources.BitcoinReceiver.OBJECT_NAME:
            api_resources.BitcoinReceiver,
        api_resources.BitcoinTransaction.OBJECT_NAME:
            api_resources.BitcoinTransaction,
        api_resources.Card.OBJECT_NAME: api_resources.Card,
        api_resources.Charge.OBJECT_NAME: api_resources.Charge,
        api_resources.CountrySpec.OBJECT_NAME: api_resources.CountrySpec,
        api_resources.Coupon.OBJECT_NAME: api_resources.Coupon,
        api_resources.Customer.OBJECT_NAME: api_resources.Customer,
        api_resources.Dispute.OBJECT_NAME: api_resources.Dispute,
        api_resources.EphemeralKey.OBJECT_NAME: api_resources.EphemeralKey,
        api_resources.Event.OBJECT_NAME: api_resources.Event,
        api_resources.FileUpload.OBJECT_NAME: api_resources.FileUpload,
        api_resources.Invoice.OBJECT_NAME: api_resources.Invoice,
        api_resources.InvoiceItem.OBJECT_NAME: api_resources.InvoiceItem,
        api_resources.LoginLink.OBJECT_NAME: api_resources.LoginLink,
        api_resources.Order.OBJECT_NAME: api_resources.Order,
        api_resources.OrderReturn.OBJECT_NAME: api_resources.OrderReturn,
        api_resources.Payout.OBJECT_NAME: api_resources.Payout,
        api_resources.Plan.OBJECT_NAME: api_resources.Plan,
        api_resources.Product.OBJECT_NAME: api_resources.Product,
        api_resources.Recipient.OBJECT_NAME: api_resources.Recipient,
        api_resources.RecipientTransfer.OBJECT_NAME:
            api_resources.RecipientTransfer,
        api_resources.Refund.OBJECT_NAME: api_resources.Refund,
        api_resources.Reversal.OBJECT_NAME: api_resources.Reversal,
        api_resources.SKU.OBJECT_NAME: api_resources.SKU,
        api_resources.Source.OBJECT_NAME: api_resources.Source,
        api_resources.Subscription.OBJECT_NAME: api_resources.Subscription,
        api_resources.SubscriptionItem.OBJECT_NAME:
            api_resources.SubscriptionItem,
        api_resources.ThreeDSecure.OBJECT_NAME: api_resources.ThreeDSecure,
        api_resources.Token.OBJECT_NAME: api_resources.Token,
        api_resources.Transfer.OBJECT_NAME: api_resources.Transfer,
    }


def convert_to_stripe_object(resp, api_key=None, stripe_version=None,
                             stripe_account=None):
    global OBJECT_CLASSES

    if len(OBJECT_CLASSES) == 0:
        load_object_classes()
    types = OBJECT_CLASSES.copy()

    if isinstance(resp, list):
        return [convert_to_stripe_object(i, api_key, stripe_version,
                                         stripe_account) for i in resp]
    elif isinstance(resp, dict) and \
            not isinstance(resp, stripe.stripe_object.StripeObject):
        resp = resp.copy()
        klass_name = resp.get('object')
        if isinstance(klass_name, basestring):
            klass = types.get(klass_name, stripe.stripe_object.StripeObject)
        else:
            klass = stripe.stripe_object.StripeObject
        return klass.construct_from(resp, api_key,
                                    stripe_version=stripe_version,
                                    stripe_account=stripe_account)
    else:
        return resp


def convert_array_to_dict(arr):
    if isinstance(arr, list):
        d = {}
        for i, value in enumerate(arr):
            d[str(i)] = value
        return d
    else:
        return arr


def populate_headers(idempotency_key):
    if idempotency_key is not None:
        return {"Idempotency-Key": idempotency_key}
    return None



import stripe
from stripe.test.helper import StripeResourceTest


class BalanceTest(StripeResourceTest):

    def test_retrieve_balance(self):
        stripe.Balance.retrieve()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/balance',
            {},
            None
        )


class BalanceTransactionTest(StripeResourceTest):

    def test_list_balance_transactions(self):
        stripe.BalanceTransaction.list()
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/balance/history',
            {}
        )



import urllib

from stripe import util
from stripe.api_resources.customer import Customer
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource


class AlipayAccount(UpdateableAPIResource, DeletableAPIResource):
    OBJECT_NAME = 'alipay_account'

    @classmethod
    def _build_instance_url(cls, customer, sid):
        token = util.utf8(sid)
        extn = urllib.quote_plus(token)
        customer = util.utf8(customer)

        base = Customer.class_url()
        owner_extn = urllib.quote_plus(customer)

        return "%s/%s/sources/%s" % (base, owner_extn, extn)

    def instance_url(self):
        return self._build_instance_url(self.customer, self.id)

    @classmethod
    def modify(cls, customer, id, **params):
        url = cls._build_instance_url(customer, id)
        return cls._modify(url, **params)

    @classmethod
    def retrieve(cls, id, api_key=None, stripe_version=None,
                 stripe_account=None, **params):
        raise NotImplementedError(
            "Can't retrieve an Alipay account without a customer ID. "
            "Use customer.sources.retrieve('alipay_account_id') instead.")



import urllib

from stripe import api_requestor, connect_api_base, error


class OAuth(object):

    @staticmethod
    def _set_client_id(params):
        if 'client_id' in params:
            return

        from stripe import client_id
        if client_id:
            params['client_id'] = client_id
            return

        raise error.AuthenticationError(
            'No client_id provided. (HINT: set your client_id using '
            '"stripe.client_id = <CLIENT-ID>"). You can find your client_ids '
            'in your Stripe dashboard at '
            'https://dashboard.stripe.com/account/applications/settings, '
            'after registering your account as a platform. See '
            'https://stripe.com/docs/connect/standalone-accounts for details, '
            'or email support@stripe.com if you have any questions.')

    @staticmethod
    def authorize_url(**params):
        path = '/oauth/authorize'
        OAuth._set_client_id(params)
        if 'response_type' not in params:
            params['response_type'] = 'code'
        query = urllib.urlencode(list(api_requestor._api_encode(params)))
        url = connect_api_base + path + '?' + query
        return url

    @staticmethod
    def token(**params):
        requestor = api_requestor.APIRequestor(api_base=connect_api_base)
        response, api_key = requestor.request(
            'post', '/oauth/token', params, None)
        return response

    @staticmethod
    def deauthorize(**params):
        requestor = api_requestor.APIRequestor(api_base=connect_api_base)
        OAuth._set_client_id(params)
        response, api_key = requestor.request(
            'post', '/oauth/deauthorize', params, None)
        return response



import datetime
import time

import stripe
from stripe.test.helper import (
    StripeResourceTest, DUMMY_PLAN
)


class SubscriptionTest(StripeResourceTest):

    def test_list_subscriptions(self):
        stripe.Subscription.all(customer="test_cus", plan=DUMMY_PLAN['id'],
                                limit=3)
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/subscriptions',
            {
                'customer': 'test_cus',
                'plan': DUMMY_PLAN['id'],
                'limit': 3,
            },
        )

    def test_retrieve_subscription(self):
        stripe.Subscription.retrieve('test_sub')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/subscriptions/test_sub',
            {},
            None
        )

    def test_create_subscription(self):
        subscription = stripe.Subscription.create(customer="test_cus",
                                                  plan=DUMMY_PLAN['id'])
        self.assertEqual({}, subscription)

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscriptions',
            {
                'customer': 'test_cus',
                'plan': DUMMY_PLAN['id']
            },
            None
        )

    def test_update_subscription(self):
        subscription = stripe.Subscription.construct_from({
            'id': 'test_sub',
            'customer': 'test_cus',
        }, 'api_key')

        trial_end_dttm = datetime.datetime.now() + datetime.timedelta(days=15)
        trial_end_int = int(time.mktime(trial_end_dttm.timetuple()))

        subscription.items = [{"id": "si", "plan": "foo"}]
        subscription.plan = DUMMY_PLAN['id']
        subscription.trial_end = trial_end_int
        subscription.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscriptions/test_sub',
            {
                'items': {
                    "0": {
                        "plan": "foo",
                        "id": "si",
                    },
                },
                'plan': DUMMY_PLAN['id'],
                'trial_end': trial_end_int
            },
            None
        )

    def test_modify_subscription(self):
        trial_end_dttm = datetime.datetime.now() + datetime.timedelta(days=15)
        trial_end_int = int(time.mktime(trial_end_dttm.timetuple()))

        subscription = stripe.Subscription.modify('test_sub',
                                                  plan=DUMMY_PLAN['id'],
                                                  trial_end=trial_end_int)
        self.assertEqual({}, subscription)

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscriptions/test_sub',
            {
                'plan': DUMMY_PLAN['id'],
                'trial_end': trial_end_int
            },
            None
        )

    def test_modify_subscription_items(self):
        stripe.Subscription.modify('test_sub',
                                   items=[{"id": "si", "plan": "foo"}])

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscriptions/test_sub',
            {
                'items': {
                    "0": {
                        "plan": "foo",
                        "id": "si",
                    },
                },
            },
            None
        )

    def test_create_subscription_items(self):
        stripe.Subscription.create(items=[{"plan": "foo", "quantity": 3}])
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscriptions',
            {
                'items': {
                    "0": {
                        "plan": "foo",
                        "quantity": 3,
                    },
                },
            },
            None
        )

    def test_delete_subscription(self):
        subscription = stripe.Subscription.construct_from({
            'id': 'test_sub',
            'customer': 'test_cus',
        }, 'api_key')

        subscription.delete()

        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/subscriptions/test_sub',
            {},
            None
        )

    def test_delete_subscription_discount(self):
        subscription = stripe.Subscription.construct_from({
            'id': 'test_sub',
            'customer': 'test_cus',
            'coupon': 'test_discount'
        }, 'api_key')

        subscription.delete_discount()

        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/subscriptions/test_sub/discount'
        )



import stripe
from stripe.test.helper import StripeResourceTest


class ReversalTest(StripeResourceTest):

    def test_fetch_reversal(self):
        transfer = stripe.Charge.construct_from({
            'id': 'tr_get',
            'reversals': {
                'object': 'list',
                'url': '/v1/transfers/tr_get/reversals',
            }
        }, 'api_key')

        transfer.reversals.retrieve("foo")

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers/tr_get/reversals/foo',
            {},
            None
        )

    def test_list_reversals(self):
        transfer = stripe.Charge.construct_from({
            'id': 'tr_list',
            'reversals': {
                'object': 'list',
                'url': '/v1/transfers/tr_list/reversals',
            }
        }, 'api_key')

        transfer.reversals.list()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers/tr_list/reversals',
            {},
            None
        )

    def test_update_transfer(self):
        reversal = stripe.Reversal.construct_from({
            'id': "rev_update",
            'transfer': "tr_update",
            'metadata': {},
        }, 'api_key')
        reversal.metadata["key"] = "value"
        reversal.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/transfers/tr_update/reversals/rev_update',
            {
                'metadata': {
                    'key': 'value',
                }
            },
            None
        )



from stripe.api_resources.abstract.api_resource import APIResource


class SingletonAPIResource(APIResource):

    @classmethod
    def retrieve(cls, **params):
        return super(SingletonAPIResource, cls).retrieve(None, **params)

    @classmethod
    def class_url(cls):
        cls_name = cls.class_name()
        return "/v1/%s" % (cls_name,)

    def instance_url(self):
        return self.class_url()



import stripe
from stripe.test.helper import (
    StripeApiTestCase, MyCreatable
)


class CreateableAPIResourceTests(StripeApiTestCase):

    def test_create(self):
        self.mock_response({
            'object': 'charge',
            'foo': 'bar',
        })

        res = MyCreatable.create()

        self.requestor_mock.request.assert_called_with(
            'post', '/v1/mycreatables', {}, None)

        self.assertTrue(isinstance(res, stripe.Charge))
        self.assertEqual('bar', res.foo)

    def test_idempotent_create(self):
        self.mock_response({
            'object': 'charge',
            'foo': 'bar',
        })

        res = MyCreatable.create(idempotency_key='foo')

        self.requestor_mock.request.assert_called_with(
            'post', '/v1/mycreatables', {}, {'Idempotency-Key': 'foo'})

        self.assertTrue(isinstance(res, stripe.Charge))
        self.assertEqual('bar', res.foo)



import stripe
from stripe.test.helper import StripeResourceTest


class ThreeDSecureTest(StripeResourceTest):

    def test_threedsecure_create(self):
        stripe.ThreeDSecure.create(
            card="tok_test",
            amount=1500,
            currency="usd",
            return_url="https://example.org/3d-secure-result"
        )

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/3d_secure',
            {
                'card': 'tok_test',
                'amount': 1500,
                'currency': 'usd',
                'return_url': 'https://example.org/3d-secure-result'
            },
            None
        )

    def test_threedsecure_retrieve(self):
        stripe.ThreeDSecure.retrieve('tdsrc_id')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/3d_secure/tdsrc_id',
            {},
            None
        )



import stripe
from stripe.test.helper import (
    StripeResourceTest, DUMMY_INVOICE_ITEM
)


class InvoiceTest(StripeResourceTest):

    def test_add_invoice_item(self):
        customer = stripe.Customer(id="cus_invoice_items")
        customer.add_invoice_item(**DUMMY_INVOICE_ITEM)

        expected = DUMMY_INVOICE_ITEM.copy()
        expected['customer'] = 'cus_invoice_items'

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/invoiceitems',
            expected,
            None,
        )

    def test_retrieve_invoice_items(self):
        customer = stripe.Customer(id="cus_get_invoice_items")
        customer.invoice_items()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/invoiceitems',
            {'customer': 'cus_get_invoice_items'},
        )

    def test_invoice_create(self):
        customer = stripe.Customer(id="cus_invoice")
        stripe.Invoice.create(customer=customer.id)

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/invoices',
            {
                'customer': 'cus_invoice',
            },
            None
        )

    def test_retrieve_customer_invoices(self):
        customer = stripe.Customer(id="cus_invoice_items")
        customer.invoices()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/invoices',
            {
                'customer': 'cus_invoice_items',
            },
        )

    def test_pay_invoice(self):
        invoice = stripe.Invoice(id="ii_pay")
        invoice.pay()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/invoices/ii_pay/pay',
            {},
            None
        )

    def test_pay_invoice_params(self):
        invoice = stripe.Invoice(id="ii_pay")
        invoice.pay(source="src_foo")

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/invoices/ii_pay/pay',
            {
                'source': 'src_foo',
            },
            None
        )

    def test_upcoming_invoice(self):
        stripe.Invoice.upcoming()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/invoices/upcoming',
            {},
        )

    def test_upcoming_invoice_subscription_items(self):
        stripe.Invoice.upcoming(subscription_items=[
            {"plan": "foo", "quantity": 3}
        ])
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/invoices/upcoming',
            {
                'subscription_items': {
                    "0": {
                        "plan": "foo",
                        "quantity": 3,
                    },
                },
            },
        )
